# -*- coding: utf-8 -*-
"""

"""
from copy import deepcopy
from typing import Dict, Tuple

from numpy import zeros, ndarray, exp, inner

from pyfem.fem.Timer import Timer
from pyfem.fem.constants import DTYPE
from pyfem.io.Material import Material
from pyfem.io.Section import Section
from pyfem.materials.BaseMaterial import BaseMaterial
from pyfem.materials.ElasticIsotropic import get_stiffness_from_young_poisson
from pyfem.utils.colors import error_style


class ViscoElasticMaxwell(BaseMaterial):
    __slots__ = BaseMaterial.__slots__ + ('E0',
                                          'E1',
                                          'E2',
                                          'E3',
                                          'TAU1',
                                          'TAU2',
                                          'TAU3',
                                          'POISSON')

    def __init__(self, material: Material, dimension: int, section: Section) -> None:
        super().__init__(material, dimension, section)
        self.allowed_section_types = ('Volume', 'PlaneStress', 'PlaneStrain')

        self.data_keys = ['E0', 'Poisson\'s ratio nu', 'E1', 'TAU1', 'E2', 'TAU2', 'E3', 'TAU3']

        if len(self.material.data) != len(self.data_keys):
            raise NotImplementedError(error_style(self.get_data_length_error_msg()))
        else:
            for i, key in enumerate(self.data_keys):
                self.data_dict[key] = material.data[i]

        self.E0: float = self.data_dict['E0']
        self.E1: float = self.data_dict['E1']
        self.E2: float = self.data_dict['E2']
        self.E3: float = self.data_dict['E3']
        self.TAU1: float = self.data_dict['TAU1']
        self.TAU2: float = self.data_dict['TAU2']
        self.TAU3: float = self.data_dict['TAU3']
        self.POISSON: float = self.data_dict['Poisson\'s ratio nu']

        self.create_tangent()

    def create_tangent(self):
        if self.section.type in self.allowed_section_types:
            self.tangent = get_stiffness_from_young_poisson(self.dimension, self.E0, self.POISSON, self.section.type)
        else:
            raise NotImplementedError(error_style(self.get_section_type_error_msg()))

    def get_tangent(self, variable: Dict[str, ndarray],
                    state_variable: Dict[str, ndarray],
                    state_variable_new: Dict[str, ndarray],
                    element_id: int,
                    igp: int,
                    ntens: int,
                    ndi: int,
                    nshr: int,
                    timer: Timer) -> Tuple[ndarray, Dict[str, ndarray]]:

        if state_variable == {}:
            state_variable['SM1'] = zeros(ntens, dtype=DTYPE)
            state_variable['SM2'] = zeros(ntens, dtype=DTYPE)
            state_variable['SM3'] = zeros(ntens, dtype=DTYPE)

        SM1OLD = deepcopy(state_variable['SM1'])
        SM2OLD = deepcopy(state_variable['SM2'])
        SM3OLD = deepcopy(state_variable['SM3'])

        # if element_id == 0 and igp == 0:
        #     print(state_variable)

        dtime = timer.dtime
        dstrain = variable['dstrain']
        strain = variable['strain']

        E0 = self.E0
        E1 = self.E1
        E2 = self.E2
        E3 = self.E3
        TAU1 = self.TAU1
        TAU2 = self.TAU2
        TAU3 = self.TAU3
        POISSON = self.POISSON

        mu0 = 0.5 * E0 / (1.0 + POISSON)
        mu1 = 0.5 * E1 / (1.0 + POISSON)
        mu2 = 0.5 * E2 / (1.0 + POISSON)
        mu3 = 0.5 * E3 / (1.0 + POISSON)

        BULK = E0 / 3 / (1.0 - 2 * POISSON)

        term1 = BULK + (4.0 * mu0) / 3.0
        term2 = BULK - (2.0 * mu0) / 3.0

        g = zeros((ntens, ntens), dtype=DTYPE)

        for i in range(ndi):
            g[i, i] = term1
        for i in range(1, ndi):
            for j in range(0, i):
                g[i, j] = term2
                g[j, i] = term2
        for i in range(ndi, ntens):
            g[i, i] = mu0

        stress = zeros(ntens, dtype=DTYPE)

        m1 = (TAU1 * mu1 - TAU1 * mu1 * exp(-dtime / TAU1)) / (mu0 * dtime)
        m2 = (TAU2 * mu2 - TAU2 * mu2 * exp(-dtime / TAU2)) / (mu0 * dtime)
        m3 = (TAU3 * mu3 - TAU3 * mu3 * exp(-dtime / TAU3)) / (mu0 * dtime)

        term3 = 1 + m1 + m2 + m3

        for i in range(ndi):
            stress[i] = inner(g[i, 0:ndi], strain[0:ndi]) + term3 * inner(g[i, 0:ndi], dstrain[0:ndi]) + SM1OLD[i] + \
                        SM2OLD[i] + SM3OLD[i]

        for i in range(ndi, ntens):
            stress[i] = g[i, i] * strain[i] + SM1OLD[i] + SM2OLD[i] + SM3OLD[i] + term3 * (g[i, i] * dstrain[i])

        # if element_id == 0 and igp == 0:
        #     print(stress)

        SM1 = zeros(ntens, dtype=DTYPE)
        SM2 = zeros(ntens, dtype=DTYPE)
        SM3 = zeros(ntens, dtype=DTYPE)

        for i in range(ndi):
            SM1[i] = SM1OLD[i] + m1 * (inner(g[i, 0:ndi], dstrain[0:ndi]))
            SM2[i] = SM2OLD[i] + m2 * (inner(g[i, 0:ndi], dstrain[0:ndi]))
            SM3[i] = SM3OLD[i] + m3 * (inner(g[i, 0:ndi], dstrain[0:ndi]))

        for i in range(ndi, ntens):
            SM1[i] = SM1OLD[i] + m1 * (g[i, i] * dstrain[i])
            SM2[i] = SM2OLD[i] + m2 * (g[i, i] * dstrain[i])
            SM3[i] = SM3OLD[i] + m3 * (g[i, i] * dstrain[i])

        for i in range(ntens):
            SM1[i] = exp(-dtime / TAU1) * SM1[i]
            SM2[i] = exp(-dtime / TAU2) * SM2[i]
            SM3[i] = exp(-dtime / TAU3) * SM3[i]

        # if element_id == 0 and igp == 0:
        #     print(SM1)

        state_variable_new['SM1'] = SM1
        state_variable_new['SM2'] = SM2
        state_variable_new['SM3'] = SM3

        ddsdde = (1 + m1 + m2 + m3) * g

        output = {'stress': stress}

        return ddsdde, output


if __name__ == "__main__":
    from pyfem.io.Properties import Properties

    props = Properties()
    props.read_file(r'..\..\..\examples\mechanical\1element\hex8\Job-1.toml')
    material_data = ViscoElasticMaxwell(props.materials[2], 3, props.sections[0])
    material_data.show()
